@@ -12,7 +12,7 @@
 
 -module(couch_rep).
 -behaviour(gen_server).
--export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2, 
+-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2,
     code_change/3]).
 
 -export([replicate/2, checkpoint/1]).
@@ -28,6 +28,7 @@
     source,
     target,
     continuous,
+    create_target,
     init_args,
     checkpoint_scheduled = nil,
 
@@ -44,7 +45,8 @@
     complete = false,
     committed_seq = 0,
 
-    stats = nil
+    stats = nil,
+    doc_ids = nil
 }).
 
 %% convenience function to do a simple replication from the shell
@@ -66,13 +68,24 @@ replicate({Props}=PostBody, UserCtx) ->
         [?MODULE]
     },
 
-    Server = start_replication_server(Replicator),
-
-    case proplists:get_value(<<"continuous">>, Props, false) of
+    case couch_util:get_value(<<"cancel">>, Props, false) of
     true ->
-        {ok, {continuous, ?l2b(BaseId)}};
+ case supervisor:terminate_child(couch_rep_sup, BaseId ++ Extension) of
+        {error, not_found} ->
+     {error, not_found};
+        ok ->
+     ok = supervisor:delete_child(couch_rep_sup, BaseId ++ Extension),
+            {ok, {cancelled, ?l2b(BaseId)}}
+ end;
     false ->
-        get_result(Server, PostBody, UserCtx)
+        Server = start_replication_server(Replicator),
+
+        case couch_util:get_value(<<"continuous">>, Props, false) of
+        true ->
+            {ok, {continuous, ?l2b(BaseId)}};
+        false ->
+            get_result(Server, PostBody, UserCtx)
+        end
     end.
 
 checkpoint(Server) ->
@@ -98,28 +111,52 @@ init(InitArgs) ->
 do_init([RepId, {PostProps}, UserCtx] = InitArgs) ->
     process_flag(trap_exit, true),
 
-    SourceProps = proplists:get_value(<<"source">>, PostProps),
-    TargetProps = proplists:get_value(<<"target">>, PostProps),
+    SourceProps = couch_util:get_value(<<"source">>, PostProps),
+    TargetProps = couch_util:get_value(<<"target">>, PostProps),
 
-    Continuous = proplists:get_value(<<"continuous">>, PostProps, false),
+    DocIds = couch_util:get_value(<<"doc_ids">>, PostProps, nil),
+    Continuous = couch_util:get_value(<<"continuous">>, PostProps, false),
+    CreateTarget = couch_util:get_value(<<"create_target">>, PostProps, false),
 
     Source = open_db(SourceProps, UserCtx),
-    Target = open_db(TargetProps, UserCtx),
-
-    SourceLog = open_replication_log(Source, RepId),
-    TargetLog = open_replication_log(Target, RepId),
+    Target = open_db(TargetProps, UserCtx, CreateTarget),
 
     SourceInfo = dbinfo(Source),
     TargetInfo = dbinfo(Target),
-    
-    {StartSeq, History} = compare_replication_logs(SourceLog, TargetLog),
 
-    {ok, ChangesFeed} =
-    couch_rep_changes_feed:start_link(self(), Source, StartSeq, PostProps),
-    {ok, MissingRevs} =
-    couch_rep_missing_revs:start_link(self(), Target, ChangesFeed, PostProps),
-    {ok, Reader} =
-    couch_rep_reader:start_link(self(), Source, MissingRevs, PostProps),
+    case DocIds of
+    List when is_list(List) ->
+        % Fast replication using only a list of doc IDs to replicate.
+        % Replication sessions, checkpoints and logs are not created
+        % since the update sequence number of the source DB is not used
+        % for determining which documents are copied into the target DB.
+        SourceLog = nil,
+        TargetLog = nil,
+
+        StartSeq = nil,
+        History = nil,
+
+        ChangesFeed = nil,
+        MissingRevs = nil,
+
+        {ok, Reader} =
+        couch_rep_reader:start_link(self(), Source, DocIds, PostProps);
+
+    _ ->
+        % Replication using the _changes API (DB sequence update numbers).
+        SourceLog = open_replication_log(Source, RepId),
+        TargetLog = open_replication_log(Target, RepId),
+    
+        {StartSeq, History} = compare_replication_logs(SourceLog, TargetLog),
+
+        {ok, ChangesFeed} =
+        couch_rep_changes_feed:start_link(self(), Source, StartSeq, PostProps),
+        {ok, MissingRevs} =
+        couch_rep_missing_revs:start_link(self(), Target, ChangesFeed, PostProps),
+        {ok, Reader} =
+        couch_rep_reader:start_link(self(), Source, MissingRevs, PostProps)
+    end,
+
     {ok, Writer} =
     couch_rep_writer:start_link(self(), Target, Reader, PostProps),
 
@@ -143,6 +180,7 @@ do_init([RepId, {PostProps}, UserCtx] = InitArgs) ->
         source = Source,
         target = Target,
         continuous = Continuous,
+        create_target = CreateTarget,
         init_args = InitArgs,
         stats = Stats,
         checkpoint_scheduled = nil,
@@ -152,8 +190,9 @@ do_init([RepId, {PostProps}, UserCtx] = InitArgs) ->
         source_log = SourceLog,
         target_log = TargetLog,
         rep_starttime = httpd_util:rfc1123_date(),
-        src_starttime = proplists:get_value(instance_start_time, SourceInfo),
-        tgt_starttime = proplists:get_value(instance_start_time, TargetInfo)
+        src_starttime = couch_util:get_value(instance_start_time, SourceInfo),
+        tgt_starttime = couch_util:get_value(instance_start_time, TargetInfo),
+        doc_ids = DocIds
     },
     {ok, State}.
 
@@ -263,17 +302,17 @@ start_replication_server(Replicator) ->
 compare_replication_logs(SrcDoc, TgtDoc) ->
     #doc{body={RepRecProps}} = SrcDoc,
     #doc{body={RepRecPropsTgt}} = TgtDoc,
-    case proplists:get_value(<<"session_id">>, RepRecProps) ==
-            proplists:get_value(<<"session_id">>, RepRecPropsTgt) of
+    case couch_util:get_value(<<"session_id">>, RepRecProps) ==
+            couch_util:get_value(<<"session_id">>, RepRecPropsTgt) of
     true ->
         % if the records have the same session id,
         % then we have a valid replication history
-        OldSeqNum = proplists:get_value(<<"source_last_seq">>, RepRecProps, 0),
-        OldHistory = proplists:get_value(<<"history">>, RepRecProps, []),
+        OldSeqNum = couch_util:get_value(<<"source_last_seq">>, RepRecProps, 0),
+        OldHistory = couch_util:get_value(<<"history">>, RepRecProps, []),
         {OldSeqNum, OldHistory};
     false ->
-        SourceHistory = proplists:get_value(<<"history">>, RepRecProps, []),
-        TargetHistory = proplists:get_value(<<"history">>, RepRecPropsTgt, []),
+        SourceHistory = couch_util:get_value(<<"history">>, RepRecProps, []),
+        TargetHistory = couch_util:get_value(<<"history">>, RepRecPropsTgt, []),
         ?LOG_INFO("Replication records differ. "
                 "Scanning histories to find a common ancestor.", []),
         ?LOG_DEBUG("Record on source:~p~nRecord on target:~p~n",
@@ -281,22 +320,22 @@ compare_replication_logs(SrcDoc, TgtDoc) ->
         compare_rep_history(SourceHistory, TargetHistory)
     end.
 
-compare_rep_history(S, T) when length(S) =:= 0 orelse length(T) =:= 0 ->
+compare_rep_history(S, T) when S =:= [] orelse T =:= [] ->
     ?LOG_INFO("no common ancestry -- performing full replication", []),
     {0, []};
 compare_rep_history([{S}|SourceRest], [{T}|TargetRest]=Target) ->
-    SourceId = proplists:get_value(<<"session_id">>, S),
+    SourceId = couch_util:get_value(<<"session_id">>, S),
     case has_session_id(SourceId, Target) of
     true ->
-        RecordSeqNum = proplists:get_value(<<"recorded_seq">>, S, 0),
+        RecordSeqNum = couch_util:get_value(<<"recorded_seq">>, S, 0),
         ?LOG_INFO("found a common replication record with source_seq ~p",
             [RecordSeqNum]),
         {RecordSeqNum, SourceRest};
     false ->
-        TargetId = proplists:get_value(<<"session_id">>, T),
+        TargetId = couch_util:get_value(<<"session_id">>, T),
         case has_session_id(TargetId, SourceRest) of
         true ->
-            RecordSeqNum = proplists:get_value(<<"recorded_seq">>, T, 0),
+            RecordSeqNum = couch_util:get_value(<<"recorded_seq">>, T, 0),
             ?LOG_INFO("found a common replication record with source_seq ~p",
                 [RecordSeqNum]),
             {RecordSeqNum, TargetRest};
@@ -305,15 +344,21 @@ compare_rep_history([{S}|SourceRest], [{T}|TargetRest]=Target) ->
         end
     end.
 
-close_db(#http_db{})->
+close_db(#http_db{}) ->
     ok;
-close_db(Db)->
+close_db(Db) ->
     couch_db:close(Db).
 
-dbname(#http_db{} = Db) ->
-    Db#http_db.url;
-dbname(Db) ->
-    Db#db.name.
+dbname(#http_db{url = Url}) ->
+    strip_password(Url);
+dbname(#db{name = Name}) ->
+    Name.
+
+strip_password(Url) ->
+    re:replace(Url,
+        "http(s)?://([^:]+):[^@]+@(.*)$",
+        "http\\1://\\2:*****@\\3",
+        [{return, list}]).
 
 dbinfo(#http_db{} = Db) ->
     {DbProps} = couch_rep_httpc:request(Db),
@@ -322,20 +367,34 @@ dbinfo(Db) ->
     {ok, Info} = couch_db:get_db_info(Db),
     Info.
 
+do_terminate(#state{doc_ids=DocIds} = State) when is_list(DocIds) ->
+    #state{
+        listeners = Listeners,
+        rep_starttime = ReplicationStartTime,
+        stats = Stats
+    } = State,
+
+    RepByDocsJson = {[
+        {<<"start_time">>, ?l2b(ReplicationStartTime)},
+        {<<"end_time">>, ?l2b(httpd_util:rfc1123_date())},
+        {<<"docs_read">>, ets:lookup_element(Stats, docs_read, 2)},
+        {<<"docs_written">>, ets:lookup_element(Stats, docs_written, 2)},
+        {<<"doc_write_failures">>,
+            ets:lookup_element(Stats, doc_write_failures, 2)}
+    ]},
+
+    terminate_cleanup(State),
+    [gen_server:reply(L, {ok, RepByDocsJson}) || L <- lists:reverse(Listeners)];
+
 do_terminate(State) ->
     #state{
         checkpoint_history = CheckpointHistory,
         committed_seq = NewSeq,
         listeners = Listeners,
         source = Source,
-        target = Target,
         continuous = Continuous,
-        stats = Stats,
         source_log = #doc{body={OldHistory}}
     } = State,
-    couch_task_status:update("Finishing"),
-    ets:delete(Stats),
-    close_db(Target),
     
     NewRepHistory = case CheckpointHistory of
     nil ->
@@ -363,41 +422,65 @@ do_terminate(State) ->
         false ->
             [gen_server:reply(R, retry) || R <- OtherListeners]
     end,
-    close_db(Source).
+    terminate_cleanup(State).
+
+terminate_cleanup(#state{source=Source, target=Target, stats=Stats}) ->
+    couch_task_status:update("Finishing"),
+    close_db(Target),
+    close_db(Source),
+    ets:delete(Stats).
 
 has_session_id(_SessionId, []) ->
     false;
 has_session_id(SessionId, [{Props} | Rest]) ->
-    case proplists:get_value(<<"session_id">>, Props, nil) of
+    case couch_util:get_value(<<"session_id">>, Props, nil) of
     SessionId ->
         true;
     _Else ->
         has_session_id(SessionId, Rest)
     end.
 
+maybe_append_options(Options, Props) ->
+    lists:foldl(fun(Option, Acc) ->
+        Acc ++
+        case couch_util:get_value(Option, Props, false) of
+        true ->
+            "+" ++ ?b2l(Option);
+        false ->
+            ""
+        end
+    end, [], Options).
+
 make_replication_id({Props}, UserCtx) ->
     %% funky algorithm to preserve backwards compatibility
     {ok, HostName} = inet:gethostname(),
     % Port = mochiweb_socket_server:get(couch_httpd, port),
-    Src = get_rep_endpoint(UserCtx, proplists:get_value(<<"source">>, Props)),
-    Tgt = get_rep_endpoint(UserCtx, proplists:get_value(<<"target">>, Props)),    
-    Base = couch_util:to_hex(erlang:md5(term_to_binary([HostName, Src, Tgt]))),
-    Extension = case proplists:get_value(<<"continuous">>, Props, false) of
-    true ->
-        "+continuous";
-    false ->
-        ""
-    end,
-    {Base, Extension}.
+    Src = get_rep_endpoint(UserCtx, couch_util:get_value(<<"source">>, Props)),
+    Tgt = get_rep_endpoint(UserCtx, couch_util:get_value(<<"target">>, Props)),
+    Base = [HostName, Src, Tgt] ++
+        case couch_util:get_value(<<"filter">>, Props) of
+        undefined ->
+            case couch_util:get_value(<<"doc_ids">>, Props) of
+            undefined ->
+                [];
+            DocIds ->
+                [DocIds]
+            end;
+        Filter ->
+            [Filter, couch_util:get_value(<<"query_params">>, Props, {[]})]
+        end,
+    Extension = maybe_append_options(
+        [<<"continuous">>, <<"create_target">>], Props),
+    {couch_util:to_hex(couch_util:md5(term_to_binary(Base))), Extension}.
 
 maybe_add_trailing_slash(Url) ->
     re:replace(Url, "[^/]$", "&/", [{return, list}]).
 
 get_rep_endpoint(_UserCtx, {Props}) ->
-    Url = maybe_add_trailing_slash(proplists:get_value(<<"url">>, Props)),
-    {BinHeaders} = proplists:get_value(<<"headers">>, Props, {[]}),
-    {Auth} = proplists:get_value(<<"auth">>, Props, {[]}),
-    case proplists:get_value(<<"oauth">>, Auth) of
+    Url = maybe_add_trailing_slash(couch_util:get_value(<<"url">>, Props)),
+    {BinHeaders} = couch_util:get_value(<<"headers">>, Props, {[]}),
+    {Auth} = couch_util:get_value(<<"auth">>, Props, {[]}),
+    case couch_util:get_value(<<"oauth">>, Auth) of
     undefined ->
         {remote, Url, [{?b2l(K),?b2l(V)} || {K,V} <- BinHeaders]};
     {OAuth} ->
@@ -432,10 +515,13 @@ open_replication_log(Db, RepId) ->
         #doc{id=DocId}
     end.
 
-open_db({Props}, _UserCtx) ->
-    Url = maybe_add_trailing_slash(proplists:get_value(<<"url">>, Props)),
-    {AuthProps} = proplists:get_value(<<"auth">>, Props, {[]}),
-    {BinHeaders} = proplists:get_value(<<"headers">>, Props, {[]}),
+open_db(Props, UserCtx) ->
+    open_db(Props, UserCtx, false).
+
+open_db({Props}, _UserCtx, CreateTarget) ->
+    Url = maybe_add_trailing_slash(couch_util:get_value(<<"url">>, Props)),
+    {AuthProps} = couch_util:get_value(<<"auth">>, Props, {[]}),
+    {BinHeaders} = couch_util:get_value(<<"headers">>, Props, {[]}),
     Headers = [{?b2l(K),?b2l(V)} || {K,V} <- BinHeaders],
     DefaultHeaders = (#http_db{})#http_db.headers,
     Db = #http_db{
@@ -443,12 +529,19 @@ open_db({Props}, _UserCtx) ->
         auth = AuthProps,
         headers = lists:ukeymerge(1, Headers, DefaultHeaders)
     },
-    couch_rep_httpc:db_exists(Db);
-open_db(<<"http://",_/binary>>=Url, _) ->
-    open_db({[{<<"url">>,Url}]}, []);
-open_db(<<"https://",_/binary>>=Url, _) ->
-    open_db({[{<<"url">>,Url}]}, []);
-open_db(<<DbName/binary>>, UserCtx) ->
+    couch_rep_httpc:db_exists(Db, CreateTarget);
+open_db(<<"http://",_/binary>>=Url, _, CreateTarget) ->
+    open_db({[{<<"url">>,Url}]}, [], CreateTarget);
+open_db(<<"https://",_/binary>>=Url, _, CreateTarget) ->
+    open_db({[{<<"url">>,Url}]}, [], CreateTarget);
+open_db(<<DbName/binary>>, UserCtx, CreateTarget) ->
+    case CreateTarget of
+    true ->
+        ok = couch_httpd:verify_is_server_admin(UserCtx),
+        couch_server:create(DbName, [{user_ctx, UserCtx}]);
+    false -> ok
+    end,
+
     case couch_db:open(DbName, [{user_ctx, UserCtx}]) of
     {ok, Db} ->
         couch_db:monitor(Db),
@@ -486,7 +579,7 @@ do_checkpoint(State) ->
     {SrcInstanceStartTime, TgtInstanceStartTime} ->
         ?LOG_INFO("recording a checkpoint for ~s -> ~s at source update_seq ~p",
             [dbname(Source), dbname(Target), NewSeqNum]),
-        SessionId = couch_util:new_uuid(),
+        SessionId = couch_uuids:random(),
         NewHistoryEntry = {[
             {<<"session_id">>, SessionId},
             {<<"start_time">>, list_to_binary(ReplicationStartTime)},
@@ -498,7 +591,7 @@ do_checkpoint(State) ->
             {<<"missing_found">>, ets:lookup_element(Stats, missing_revs, 2)},
             {<<"docs_read">>, ets:lookup_element(Stats, docs_read, 2)},
             {<<"docs_written">>, ets:lookup_element(Stats, docs_written, 2)},
-            {<<"doc_write_failures">>, 
+            {<<"doc_write_failures">>,
                 ets:lookup_element(Stats, doc_write_failures, 2)}
         ]},
         % limit history to 50 entries
@@ -533,11 +626,13 @@ do_checkpoint(State) ->
             reader = Reader,
             writer = Writer
         } = State,
-        Pids = [CF, MR, Reader, Writer],
+        Pids = [Writer, Reader, MR, CF],
         [unlink(Pid) || Pid <- Pids],
         [exit(Pid, shutdown) || Pid <- Pids],
+        close_db(Target),
+        close_db(Source),
         {ok, NewState} = init(State#state.init_args),
-        NewState
+        NewState#state{listeners=State#state.listeners}
     end.
 
 commit_to_both(Source, Target, RequiredSeq) ->
@@ -558,17 +653,17 @@ commit_to_both(Source, Target, RequiredSeq) ->
     end,
     {SourceStartTime, TargetStartTime}.
     
-ensure_full_commit(#http_db{} = Target) ->
+ensure_full_commit(#http_db{headers = Headers} = Target) ->
     Req = Target#http_db{
         resource = "_ensure_full_commit",
         method = post,
-        body = true
+        headers = [{"content-type", "application/json"} | Headers]
     },
     {ResultProps} = couch_rep_httpc:request(Req),
-    true = proplists:get_value(<<"ok">>, ResultProps),
-    proplists:get_value(<<"instance_start_time">>, ResultProps);
+    true = couch_util:get_value(<<"ok">>, ResultProps),
+    couch_util:get_value(<<"instance_start_time">>, ResultProps);
 ensure_full_commit(Target) ->
-    {ok, NewDb} = couch_db:open(Target#db.name, []),
+    {ok, NewDb} = couch_db:open_int(Target#db.name, []),
     UpdateSeq = couch_db:get_update_seq(Target),
     CommitSeq = couch_db:get_committed_update_seq(NewDb),
     InstanceStartTime = NewDb#db.instance_start_time,
@@ -583,20 +678,20 @@ ensure_full_commit(Target) ->
         InstanceStartTime
     end.
 
-ensure_full_commit(#http_db{} = Source, RequiredSeq) ->
+ensure_full_commit(#http_db{headers = Headers} = Source, RequiredSeq) ->
     Req = Source#http_db{
         resource = "_ensure_full_commit",
         method = post,
-        body = true,
-        qs = [{seq, RequiredSeq}]
+        qs = [{seq, RequiredSeq}],
+        headers = [{"content-type", "application/json"} | Headers]
     },
     {ResultProps} = couch_rep_httpc:request(Req),
-    case proplists:get_value(<<"ok">>, ResultProps) of
+    case couch_util:get_value(<<"ok">>, ResultProps) of
     true ->
-        proplists:get_value(<<"instance_start_time">>, ResultProps);
+        couch_util:get_value(<<"instance_start_time">>, ResultProps);
     undefined -> nil end;
 ensure_full_commit(Source, RequiredSeq) ->
-    {ok, NewDb} = couch_db:open(Source#db.name, []),
+    {ok, NewDb} = couch_db:open_int(Source#db.name, []),
     CommitSeq = couch_db:get_committed_update_seq(NewDb),
     InstanceStartTime = NewDb#db.instance_start_time,
     couch_db:close(NewDb),
@@ -618,7 +713,7 @@ update_local_doc(#http_db{} = Db, #doc{id=DocId} = Doc) ->
         headers = [{"x-couch-full-commit", "false"} | Db#http_db.headers]
     },
     {ResponseMembers} = couch_rep_httpc:request(Req),
-    Rev = proplists:get_value(<<"rev">>, ResponseMembers),
+    Rev = couch_util:get_value(<<"rev">>, ResponseMembers),
     couch_doc:parse_rev(Rev);
 update_local_doc(Db, Doc) ->
     {ok, Result} = couch_db:update_doc(Db, Doc, [delay_commit]),
@@ -627,7 +722,7 @@ update_local_doc(Db, Doc) ->
 up_to_date(#http_db{}, _Seq) ->
     true;
 up_to_date(Source, Seq) ->
-    {ok, NewDb} = couch_db:open(Source#db.name, []),
+    {ok, NewDb} = couch_db:open_int(Source#db.name, []),
     T = NewDb#db.update_seq == Seq,
     couch_db:close(NewDb),
     T.
